"""
Visualization methods.

| Copyright 2017-2024, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""

# For backwards-compatibility with older versions of plugins like
# https://github.com/voxel51/fiftyone-plugins/blob/5c800f1ded53c285f8e17f37e1ad9b2472fa93e7/plugins/brain/__init__.py#L25
from fiftyone.brain.visualization import (
    Visualization,
    UMAPVisualization,
    TSNEVisualization,
    PCAVisualization,
    ManualVisualization,
)
