"""
Internal FiftyOne Brain package.

Contains all non-public code powering the ``fiftyone.brain`` public namespace.

| Copyright 2017-2024, Voxel51, Inc.
| `voxel51.com <https://voxel51.com/>`_
|
"""
